/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef CUBELIB_POP_HYBRID_THREAD_EFFICIENCY_TEST_H
#define CUBELIB_POP_HYBRID_THREAD_EFFICIENCY_TEST_H

#include <string>
#include "PerformanceTest.h"
#include "POPHybridOMPRegionEfficiencyTest.h"
#include "POPHybridAmdahlTest.h"

namespace hybanalysis
{
class POPHybridParallelEfficiencyTest;
class POPHybridThreadEfficiencyTest : public popcalculation::PerformanceTest
{
    friend  class POPHybridParallelEfficiencyTest;
private:
    POPHybridAmdahlTest*              amdahl_eff;
    POPHybridOmpRegionEfficiencyTest* omp_region_eff;

    cube::Metric* pop_avg_comp;
    cube::Metric* pop_max_omp_comp;
    cube::Metric* pop_avg_ser_comp;

    cube::list_of_metrics lavg_omp_metrics;
    cube::list_of_metrics lavg_ser_metrics;

    void
    calculate();

    void
    calculate( const cube::list_of_cnodes& );

protected:

    virtual
    const std::string&
    getCommentText() const;

    inline
    virtual
    std::string
    getHelpUrl() const
    {
        std::string active_text =
            "Thread Efficiency considers two sources of inefficiency: \n"
            "\tSerial computation on the master outside OpenMP, i.e. reflects Amdahl's law \n"
            "\tInefficiencies within threads, e.g. serialisation across threads \n"
            "These two can be measured with Amdahl's Efficeincy and OpenMP region Efficiency respectively. \n"
            "Thread Efficeincy can be computed directly or as a product of these two sub-metrics: \n"
            "\tThread Efficiency = avg( computation time)/(avg( time in OpenMP ) + avg( serial computation ) )\n"
            "\t                  = Amdahl's Efficiency x OpenMP Region Efficiency\n"
            "Where average time in OpenMP and average serial computation are computed as weighted arithmetic mean.\n"
            "If number of threads is equal across processes average time in OpenMP and \n"
            "average serial computation can be computed as ordinary arithmetic mean. ";
        std::string not_active_text =
            "Thread Efficiency metric is a basic POP metric and is available for every Score-P/Scalasca measurement. \n"
            "If Cube Report was produced by another tool than Score-P/Scalasca, it might have missing metric Time.\n"
            "In this case POP analysis is not possible.";

        return active_text;
    }

    virtual
    void
    adjustForTest( cube::CubeProxy* cube ) const;


public:
    POPHybridThreadEfficiencyTest(     cube::CubeProxy*,
                                       POPHybridAmdahlTest*              amdahl_eff,
                                       POPHybridOmpRegionEfficiencyTest* omp_region_eff
                                       );

    virtual
    ~POPHybridThreadEfficiencyTest()
    {
    };
// ------ overview tests ---------

    void
    applyCnode( const cube::list_of_cnodes& cnodes,
                const bool                  direct_calculation = false   );


    bool
    isActive() const;

    bool
    isIssue() const;

    virtual
    std::list<PerformanceTest*>
    getPrereqs()
    {
        std::list<PerformanceTest*> prereqs;
        if ( amdahl_eff != nullptr && omp_region_eff != nullptr )
        {
            prereqs.push_back( amdahl_eff );
            prereqs.push_back( omp_region_eff );
        }
        return prereqs;
    }
};
};
#endif // ADVISER_RATING_WIDGET_H
